from dataclasses import dataclass
from .utils import Dictable
from .enumerators import SlotType, IntentType, IssueType, IssueStatus, ResponseType, QuestionMode
from .dataclasses import SlotVal, Issue
from typing import Any, List, Tuple, Union, Dict
from copy import deepcopy
from datetime import datetime


@Dictable.decorator
@dataclass
class SlotDef(Dictable):
    name: str
    type: SlotType
    entities: List[str]
    intent: List[Tuple[str, Any]]
    options: Union[None, List[Any]]
    default: Union[None, List[Tuple[str, Any]]]
    defaultQuestionCode: str
    canUsePrev: bool


@Dictable.decorator
@dataclass
class IntentDefinition(Dictable):
    type: IntentType
    name: str
    responseCodes: List[str]


@Dictable.decorator
@dataclass
class ResolveDefinition(Dictable):
    slotValues: List[SlotVal]
    responseCodes: List[str]


@Dictable.decorator
@dataclass
class ResponseDefinition(Dictable):
    type: ResponseType
    code: str
    appId: str
    slots: List[str]
    slotVariations: Dict[List[str], str]
    slotType: SlotType
    questionMode: QuestionMode
    relatedSlot: Union[None, str]
    slotFillValue: Union[None, Any]
    relatedIssues: Union[None, str]


@Dictable.decorator
@dataclass
class IssueDefinition(Dictable):
    type: IssueType
    name: str
    slotNames: List[str]
    possibleResolves: List[ResolveDefinition]
    relatedQuestion: Union[None, str]

    def build_issue(self, slotDefs: List[SlotDef], id=None, context=None, userMessage=None):
        filled_slots = {}
        for slotDef in slotDefs:
            slotName = slotDef.name
            if slotDef.canUsePrev:
                if context is not None:
                    for prev_issue in (context.activeIssues + context.inactiveIssues):
                        if slotName in prev_issue.slotValues:
                            filled_slots[slotName] = deepcopy(
                                prev_issue.slotValues[slotName]
                            )
                            break
            if userMessage is not None:
                for entity_name in slotDef.entities:
                    entity_val = userMessage.get_entity_val(entity_name)
                    if entity_val is not None:
                        filled_slots[slotName] = SlotVal(
                            slotName, entity_name, entity_val, None)
                        break
                else:
                    for intent, value in slotDef.intent:
                        if intent in userMessage.intent.positives:
                            filled_slots[slotName] = SlotVal(
                                slotName, None, value, None)
                            break

        return Issue(
            id=Issue.new_id() if id is None else id,
            type=self.type,
            name=self.name,
            created=datetime.now(),
            assignees=[],
            status=IssueStatus.ACTIVE,
            resolveAttempts=[],
            slotValues=filled_slots,
            expectingAnswers=[],
        )


@Dictable.decorator
@dataclass
class EventDefinition(Dictable):
    name: str
    issue: Union[None, str]
    codes: List[str]
    onFail: str



import redis
from json import dumps, loads
import requests
from typing import Any, List, Union
from enum import Enum
from typing_utils import get_origin
from datetime import datetime
from dataclasses import _FIELDS  # type:ignore
from uuid import uuid1
DATETIME_FORMAT = "%Y-%m-%dZ%H:%M:%S.%f"


class Cache(object):
    def __init__(self, config):
        self.redis = redis.Redis(host=config["host"], port=config["port"], db=0)
        self.ttl = config["ttl"]
        self.prefix = config["prefix"] + ":"

    def __contains__(self, key):
        return self.redis.exists(self.prefix + str(key))

    def __getitem__(self, key):
        val = self.redis.get(self.prefix + str(key))
        if val is None:
            raise ValueError("Cache Key Not Found", key)
        return loads(val)

    def __setitem__(self, key, value):
        self.redis.set(self.prefix + str(key), dumps(value), ex=self.ttl)

    def __delitem__(self, key):
        return self.redis.delete(self.prefix + str(key))

    def clear(self):
        return self.redis.flushall()


class Module:
    def __init__(self, config):
        self.name = config["name"]
        if "cache" in config:
            self.cache = Cache(config["cache"])
        self.use_cache = "cache" in config
        self.address = f"http://{config['host']}:{config['port']}/"
        self.config = config

    def post(self, endpoint, *args):
        try:
            resp = requests.post(self.address + endpoint, json={"__args__": args})
            return resp.json()["data"]
        except Exception as e:
            print(f"ERROR: {self.name}", args, flush=True)
            print(e, flush=True)


def _fill(dict, key, default, builder=None, isList=False):
    try:
        val = dict.get(key, default)
        if val is not None:
            if builder is not None:
                if isList:
                    val = [builder(v) for v in val]
                else:
                    val = builder(val)
        dict[key] = val
    except Exception as e:
        print("FILL Error:", key, dict, default)
        raise e


class Dictable:
    filler_args: List

    @staticmethod
    def decorator(cls_def):
        def type_matcher(type):
            if get_origin(type) == list:
                _, builder, _ = type_matcher(type.__args__[0])
                return [], builder, True
            elif get_origin(type) == dict:
                _, builder, _ = type_matcher(type.__args__[1])
                return {}, lambda d: {k: builder(v) for k, v in d.items()}, False
            elif get_origin(type) == Union:
                _, builder, _ = type_matcher(type.__args__[1])
                return None, builder, False
            elif get_origin(type) == tuple:
                return (), lambda x: x, False
            elif type == Any:
                return None, lambda x: x, False
            elif issubclass(type, Enum):
                return "", type, False
            elif issubclass(type, Dictable):
                return {}, type.from_dict, False
            elif type == datetime:
                return (
                    datetime.strftime(datetime.now(), DATETIME_FORMAT),
                    lambda txt: datetime.strptime(txt, DATETIME_FORMAT),
                    False,
                )
            else:
                return None, lambda x: x, None

        cls_def.filler_args = []
        attributes = getattr(cls_def.__mro__[0], _FIELDS, None)
        if attributes is None:
            raise ValueError
        for name, attribute in attributes.items():
            default, builder, isList = type_matcher(attribute.type)
            if isList is not None:
                cls_def.filler_args.append((name, default, builder, isList))

        cls_def.__hash__ = Dictable.__hash__
        return cls_def

    def to_dict(self):
        def type_matcher(v):
            if isinstance(v, Dictable):
                return v.to_dict()
            elif isinstance(v, list):
                return [type_matcher(v_) for v_ in v]
            elif isinstance(v, dict):
                return {k: type_matcher(v_) for k, v_ in v.items()}
            elif isinstance(v, datetime):
                return datetime.strftime(v, DATETIME_FORMAT)
            elif isinstance(v, Enum):
                return v.value
            else:
                return v

        return {k: type_matcher(v) for k, v in self.__dict__.items()}

    @classmethod
    def from_dict(cls, dictionary):
        for name, default, builder, isList in cls.filler_args:
            _fill(dictionary, name, default, builder, isList)
        return cls(**dictionary)

    def __hash__(self) -> int:
        return hash(dumps(self.to_dict(), sort_keys=True))

    @staticmethod
    def new_id():
        return str(uuid1())



from enum import Enum
from .utils import Dictable


class IssueStatus(Enum):
    ACTIVE = "active"
    RESOLVED = "resolved"
    UNRESOLVED = "unresolved"
    FORWARDED = "forwarded"


class SlotType(Enum):
    STRING = "string"
    BOOL = "bool"
    CLASS = "class"
    NUMERIC = "numeric"
    DATE = "date"
    API = "api"
    ANY = "any"


class CreatorType(Enum):
    BOT = "bot"
    HUMAN = "human"


class IssueType(Enum):
    COMPLAINT = "complaint"
    INFO = "info"
    PROBLEM = "problem"
    OTHER = "other"


class IntentType(Enum):
    ISSUE = "issue"
    DEFINITE = "definite"
    INDEFINITE = "indefinite"


class ResponseType(Enum):
    STRING = "question"
    BOOL = "text"


class QuestionMode(Enum):
    ISSUE_CREATE = "issueCreate"
    ISSUE_CLOSE = "issueClose"
    SLOT = "slot"
    SLOT_FILL = "slotFill"
    VARIABLE = "variable"


class ContextOperationType(Enum):
    REMOVE_NOTIFICATION = "removeNotification"
    ADD_NOTIFICATION = "addNotification"
    REMOVE_QUESTION = "removeQuestion"
    ADD_QUESTION = "addQuestion"
    CREATE_ISSUE = "createIssue"
    CREATE_SLOT = "createSlot"
    REMOVE_SLOT = "removeSlot"
    REMOVE_EMAIL = "removeEmail"
    ADD_EMAIL = "addEmail"
    REMOVE_PHONE = "removePhone"
    ADD_PHONE = "addPhone"
    REGISTER_USER_ID = "registerUserId"
    CREATE_RESOLVE = "createResolve"
    UPDATE_ISSUE_STATUS = "updateIssueStatus"
    ADD_ASSIGNEE = "addAssignee"
    REMOVE_ASSIGNEE = "removeAssignee"
    UPDATE_SLOT_VALUE = "updateSlotValue"
    UPDATE_SLOT_DEFAULT_VALUE = "updateSlotDefaultValue"
    UPDATE_RESOLVE = "updateResolve"


class ContextOperation(Dictable):
    def __init__(self, type: ContextOperationType, **kwargs):
        self.type = type
        self.args = kwargs

    def get(self, name):
        return self.args[name]

    def has(self, name):
        return name in self.args

    def __repr__(self):
        return f"Context Operation: {self.type}\n\t" + str(self.args)

    @classmethod
    def list_diff(cls, list_a, list_b, not_in_b, not_in_a, elem, **kwargs):
        ops = []
        for elem_a in list_a:
            if elem_a not in list_b:
                kwargs[elem] = elem_a
                ops.append(cls(not_in_b, **kwargs))
        for elem_b in list_b:
            if elem_b not in list_a:
                kwargs[elem] = elem_b
                ops.append(cls(not_in_a, **kwargs))
        return ops

    @classmethod
    def dict_diff(cls, dict_a, dict_b, not_in_b, not_in_a, elem, **kwargs):
        ops = []
        for key_a, val_a in dict_a.items():
            if key_a not in dict_b:
                kwargs[elem] = val_a
                ops.append(cls(not_in_b, **kwargs))
            else:
                ops += val_a.diff(dict_b[key_a], **kwargs)
        for key_b, val_b in dict_b.items():
            if key_b not in dict_a:
                kwargs[elem] = val_b
                ops.append(cls(not_in_a, **kwargs))
        return ops



from dataclasses import dataclass
from typing import Dict, Any, List, Tuple, Union
from datetime import datetime
from .enumerators import (
    CreatorType, SlotType, QuestionMode,
    IssueType, IssueStatus
)
from .utils import Dictable


@Dictable.decorator
@dataclass
class Text(Dictable):
    mutations: Dict[str, str]

    def __init__(self, mutations: Union[str, Dict[str, str]]):
        if isinstance(mutations, str):
            self.mutations = {"original": mutations}
        else:
            self.mutations = mutations

    def add_mutation(self, key, mutation):
        self.mutations[key] = mutation

    def get(self, key="original") -> str:
        return self.mutations[key] if key in self.mutations else ""


@Dictable.decorator
@dataclass
class Entity(Dictable):
    name: str
    input_type: str
    val: str
    sidx: int
    eidx: int
    confidence: float


@Dictable.decorator
@dataclass
class Intent(Dictable):
    chosen: str
    scores: Dict[str, float]
    positives: List[str]
    negatives: List[str]


@Dictable.decorator
@dataclass
class ResolveAttempt(Dictable):
    id: str
    created: datetime
    accepted: Union[None, bool]
    creator: CreatorType
    responseCodes: List[str]

    def is_equal(self, other):
        return set(self.responseCodes) == set(other.responseCodes)


@Dictable.decorator
@dataclass
class Question(Dictable):
    code: str
    type: SlotType
    mode: QuestionMode
    relatedSlotName: Union[None, str]
    slotFillValue: Union[None, List[Tuple[str, Any]]]
    relatedIssueName: Union[None, str]
    resolveId: Union[None, str]


@Dictable.decorator
@dataclass
class SlotVal(Dictable):
    name: str
    takenFromEntity: Union[None, str]
    value: Union[None, Any]  # TODO what if user dont want to give info
    default: Union[None, Any]

    def update_value(self, msg):
        if self.takenFromEntity is not None:
            self.value = msg.get_entity_val(self.takenFromEntity)


@Dictable.decorator
@dataclass
class Issue(Dictable):
    id: str
    type: IssueType
    name: str
    created: datetime
    assignees: List[str]
    status: IssueStatus
    resolveAttempts: List[ResolveAttempt]
    slotValues: Dict[str, SlotVal]
    expectingAnswers: List[Question]


@Dictable.decorator
@dataclass
class UserSignature(Dictable):
    userId: Union[None, str]
    email: List[str]
    phone: List[str]


@Dictable.decorator
@dataclass
class Response(Dictable):
    code: str
    appId: str
    relatedSlot: Union[None, str]
    relatedIssueId: Union[None, str]


@Dictable.decorator
@dataclass
class Context(Dictable):
    userSignature: UserSignature
    variables: Dict[str, SlotVal]
    activeIssues: List[Issue]
    inactiveIssues: List[Issue]
    impendingQuestions: List[Question]
    impendingNotifications: List[str]
    lastResponseCodes: List[str]
    appId: str

    def apply_last_responses(self, responses: List[Response]):
        self.lastResponseCodes = [resp.code for resp in responses]

    def get_issue(self, issueId) -> Union[None, Issue]:
        for issue in self.activeIssues + self.inactiveIssues:
            if issue.id == issueId:
                return issue
        return None

    def get_resolve(self, resolveId) -> Union[None, ResolveAttempt]:
        for issue in self.activeIssues + self.inactiveIssues:
            for resolve in issue.resolveAttempts:
                if resolve.id == resolveId:
                    return resolve
        return None


@Dictable.decorator
@dataclass
class UserMessage(Dictable):
    user_id: str
    app_id: str
    channel_id: str
    text: Union[None, Text]
    intent: Union[None, Intent]
    entities: List[Entity]
    context: Union[None, Context]

    def select(self, key):
        args = key.split(":")
        if args[0] == "text":
            return self.text.get(args[1])  # type:ignore
        elif args[0] == "intent":
            return self.intent.chosen  # type:ignore
        elif args[0] == "entity":
            for e in self.entities:  # type:ignore
                if e.name == args[1]:
                    return e.val
        return ""

    def get_entity_val(self, name, default=None):
        if self.entities is not None:
            for e in self.entities:
                if e.name == name:
                    return e.val
        return default


@Dictable.decorator
@dataclass
class BotMessage(Dictable):
    user_id: str
    app_id: str
    channel_id: str
    code: Union[None, str]
    text: Union[None, str]
    buttons: List[str]
    misc: Any


@Dictable.decorator
@dataclass
class Event(Dictable):
    name: str
    responseCode: Union[None, str]
    issueQuestions: List[Question]
    responseVars: Any
    isFailed: bool



from typing import List, Union
from copy import deepcopy

from .utils import Dictable
from .enumerators import ContextOperationType, IssueStatus
from .dataclasses import (
    ResolveAttempt, SlotVal, Issue, UserSignature, Context, UserMessage
)


class Operation(Dictable):
    def __init__(self, type: ContextOperationType, **kwargs):
        self.type = type
        self.args = kwargs

    def get(self, name):
        return self.args[name]

    def has(self, name):
        return name in self.args

    def __repr__(self):
        return f"Context Operation: {self.type}\n\t" + str(self.args)

    @classmethod
    def list_diff(cls, list_a, list_b, not_in_b, not_in_a, elem, **kwargs):
        ops = []
        for elem_a in list_a:
            if elem_a not in list_b:
                kwargs[elem] = elem_a
                ops.append(cls(not_in_b, **kwargs))
        for elem_b in list_b:
            if elem_b not in list_a:
                kwargs[elem] = elem_b
                ops.append(cls(not_in_a, **kwargs))
        return ops

    @classmethod
    def dict_diff(cls, dict_a, dict_b, not_in_b, not_in_a, elem, **kwargs):
        ops = []
        for key_a, val_a in dict_a.items():
            if key_a not in dict_b:
                _kwargs = deepcopy(kwargs)
                _kwargs[elem] = val_a
                ops.append(cls(not_in_b, **_kwargs))
            else:
                ops += diff(val_a, dict_b[key_a], **kwargs)
        for key_b, val_b in dict_b.items():
            if key_b not in dict_a:
                _kwargs = deepcopy(kwargs)
                _kwargs[elem] = val_b
                ops.append(cls(not_in_a, **_kwargs))
        return ops


def diff(a, b, **kwargs) -> List[Operation]:
    def get_issue_id(issue):
        if kwargs.get("test", False):
            return issue.name
        else:
            return issue.id
    if isinstance(a, ResolveAttempt):
        ops = []
        if a.accepted != b.accepted:
            ops.append(
                Operation(
                    ContextOperationType.UPDATE_RESOLVE,
                    issueId=kwargs["issueId"],
                    resolveId=a.id,
                    accepted=b.accepted,
                )
            )
        return ops
    elif isinstance(a, SlotVal):
        ops = []
        if a.value != b.value:
            ops.append(
                Operation(
                    ContextOperationType.UPDATE_SLOT_VALUE,
                    slot=a.name,
                    value=b.value,
                    issueId=kwargs["issueId"],
                )
            )
        if a.default != b.default:
            ops.append(
                Operation(
                    ContextOperationType.UPDATE_SLOT_DEFAULT_VALUE,
                    slot=a.name,
                    value=b.default,
                    issueId=kwargs["issueId"],
                )
            )
        return ops
    if isinstance(a, Issue):
        ops = []
        ops += Operation.list_diff(
            a.assignees,
            b.assignees,
            ContextOperationType.ADD_ASSIGNEE,
            ContextOperationType.REMOVE_ASSIGNEE,
            "assignee",
        )
        if a.status != b.status:
            ops.append(
                Operation(
                    ContextOperationType.UPDATE_ISSUE_STATUS,
                    issueId=a.id,
                    status=b.status,
                )
            )

        for other_resolve in b.resolveAttempts:
            for my_resolve in a.resolveAttempts:
                if my_resolve.is_equal(other_resolve):
                    break
            else:
                ops.append(
                    Operation(
                        ContextOperationType.CREATE_RESOLVE,
                        resolve=other_resolve,
                        issueId=a.id,
                    )
                )
        for resolve in a.resolveAttempts:
            for other_resolve in b.resolveAttempts:
                if resolve.id == other_resolve.id:
                    ops += diff(resolve, other_resolve, issueId=a.id)
                    break
            else:
                if not kwargs["silent"]:
                    raise ValueError(
                        "Resolve not found in new context",
                        {
                            "resolve": resolve.to_dict(),
                            "current": a.to_dict(),
                            "next": b.to_dict(),
                        },
                    )

        ops += Operation.dict_diff(
            a.slotValues,
            b.slotValues,
            ContextOperationType.REMOVE_SLOT,
            ContextOperationType.CREATE_SLOT,
            "slot",
            issueId=a.id,
        )
        ops.extend(
            Operation.list_diff(
                a.expectingAnswers,
                b.expectingAnswers,
                ContextOperationType.REMOVE_QUESTION,
                ContextOperationType.ADD_QUESTION,
                "question",
                issueId=a.id,
            )
        )
        return ops
    if isinstance(b, UserSignature):
        ops = []
        if a.userId is None and b.userId is not None:
            ops.append(
                Operation(
                    ContextOperationType.REGISTER_USER_ID, userId=b.userId
                )
            )
        elif a.userId != b.userId:
            raise ValueError(
                "User ids does not match",
                {"current user id": a.userId, "new user id": b.userId},
            )
        ops.extend(
            Operation.list_diff(
                a.email,
                b.email,
                ContextOperationType.REMOVE_EMAIL,
                ContextOperationType.ADD_EMAIL,
                "email",
            )
        )
        ops.extend(
            Operation.list_diff(
                a.phone,
                b.phone,
                ContextOperationType.REMOVE_PHONE,
                ContextOperationType.ADD_PHONE,
                "phone",
            )
        )
        return ops
    if isinstance(b, Context):
        ops = []
        ops += diff(a.userSignature, b.userSignature)

        ops += Operation.dict_diff(
            a.variables,
            b.variables,
            ContextOperationType.REMOVE_SLOT,
            ContextOperationType.CREATE_SLOT,
            "slot",
            issueId=None,
        )

        active_issue_ids = [get_issue_id(i) for i in a.activeIssues]
        for other_issue in b.activeIssues:
            if get_issue_id(other_issue) not in active_issue_ids:
                ops.append(
                    Operation(
                        ContextOperationType.CREATE_ISSUE, issue=other_issue
                    )
                )
                ops += diff(Issue.from_dict({
                    "name": other_issue.name,
                    "id": other_issue.id,
                    "type": other_issue.type.value,
                    "status": other_issue.status.value
                }), other_issue)
        for issue in a.activeIssues:
            for other_issue in b.activeIssues:
                if get_issue_id(issue) == get_issue_id(other_issue):
                    ops += diff(issue, other_issue)
                    break
            else:
                for other_issue in b.inactiveIssues:
                    if get_issue_id(issue) == get_issue_id(other_issue):
                        ops += diff(issue, other_issue)
                        break
                else:
                    raise ValueError(
                        "Issue not found in new context",
                        {
                            "issue": issue.id,
                            "current": a.to_dict(),
                            "next": b.to_dict(),
                        },
                    )
        ops.extend(
            Operation.list_diff(
                a.impendingQuestions,
                b.impendingQuestions,
                ContextOperationType.REMOVE_QUESTION,
                ContextOperationType.ADD_QUESTION,
                "question",
                issueId=None,
            )
        )
        ops.extend(
            Operation.list_diff(
                a.impendingNotifications,
                b.impendingNotifications,
                ContextOperationType.REMOVE_NOTIFICATION,
                ContextOperationType.ADD_NOTIFICATION,
                "notification",
            )
        )
        return ops
    else:
        raise TypeError("Invalid Type For Difference Calculation", a, b)


def apply(a: Context, ops: List[Operation], msg: Union[None, UserMessage] = None):
    def _is(type):
        return op.type == type

    for op in ops:
        if _is(ContextOperationType.REMOVE_NOTIFICATION):
            a.impendingNotifications.remove(op.get("notification"))
        elif _is(ContextOperationType.ADD_NOTIFICATION):
            a.impendingNotifications.append(op.get("notification"))
        elif _is(ContextOperationType.REMOVE_QUESTION):
            if op.get("issueId") is None:
                for q in a.impendingQuestions:
                    if q.code == op.get("question").code:
                        a.impendingQuestions.remove(q)
            else:
                issue = a.get_issue(op.get("issueId"))
                if issue is not None:
                    issue.expectingAnswers.remove(op.get("question"))
        elif _is(ContextOperationType.ADD_QUESTION):
            if op.get("issueId") is None:
                a.impendingQuestions.append(op.get("question"))
            else:
                issue = a.get_issue(op.get("issueId"))
                if issue is not None:
                    issue.expectingAnswers.append(op.get("question"))
        elif _is(ContextOperationType.CREATE_ISSUE):
            a.activeIssues.append(op.get("issue"))
        elif _is(ContextOperationType.CREATE_SLOT):
            slot: SlotVal = op.get("slot")
            if msg is not None:
                slot.update_value(msg)
            if op.get("issueId") is None:
                a.variables[op.get("slot").name] = op.get("slot")
            else:
                issue = a.get_issue(op.get("issueId"))
                if issue is not None:
                    issue.slotValues[op.get("slot").name] = op.get("slot")
        elif _is(ContextOperationType.REMOVE_SLOT):
            if op.get("issueId") is None:
                del a.variables[op.get("slot").name]
            else:
                issue = a.get_issue(op.get("issueId"))
                if issue is not None:
                    del issue.slotValues[op.get("slot").name]
        elif _is(ContextOperationType.REMOVE_EMAIL):
            a.userSignature.email.remove(op.get("email"))
        elif _is(ContextOperationType.ADD_EMAIL):
            a.userSignature.email.append(op.get("email"))
        elif _is(ContextOperationType.REMOVE_PHONE):
            a.userSignature.phone.remove(op.get("phone"))
        elif _is(ContextOperationType.ADD_PHONE):
            a.userSignature.phone.append(op.get("phone"))
        elif _is(ContextOperationType.REGISTER_USER_ID):
            a.userSignature.userId = op.get("userId")
        elif _is(ContextOperationType.CREATE_RESOLVE):
            issue = a.get_issue(op.get("issueId"))
            if issue is not None:
                issue.resolveAttempts.append(op.get("resolve"))
        elif _is(ContextOperationType.UPDATE_ISSUE_STATUS):
            issue = a.get_issue(op.get("issueId"))
            if issue is not None:
                prev = issue.status
                new = op.get("status")
                issue.status = new
                if prev == IssueStatus.ACTIVE and new != IssueStatus.ACTIVE:
                    a.activeIssues.remove(issue)
                    a.inactiveIssues.append(issue)
                elif prev != IssueStatus.ACTIVE and new == IssueStatus.ACTIVE:
                    a.activeIssues.append(issue)
                    a.inactiveIssues.remove(issue)
        elif _is(ContextOperationType.ADD_ASSIGNEE):
            issue = a.get_issue(op.get("issueId"))
            if issue is not None:
                issue.assignees.append(op.get("assignee"))
        elif _is(ContextOperationType.ADD_ASSIGNEE):
            issue = a.get_issue(op.get("issueId"))
            if issue is not None:
                issue.assignees.remove(op.get("assignee"))
        elif _is(ContextOperationType.UPDATE_RESOLVE):
            resolve = a.get_resolve(op.get("resolveId"))
            if resolve is not None:
                resolve.accepted = op.get("accepted")
        elif _is(ContextOperationType.UPDATE_SLOT_VALUE):
            slotName = op.get("slot")
            issueId = op.get("issueId")
            if issueId is None:
                a.variables[slotName].value = op.get("value")
                if msg is not None:
                    a.variables[slotName].update_value(msg)
                a.variables[slotName].value = op.get("value")
            else:
                for issue in a.activeIssues + a.inactiveIssues:
                    if issue.id == issueId:
                        issue.slotValues[slotName].value = op.get("value")
                        if msg is not None:
                            issue.slotValues[slotName].update_value(msg)
        elif _is(ContextOperationType.UPDATE_SLOT_DEFAULT_VALUE):
            slotName = op.get("slot")
            issueId = op.get("issueId")
            if issueId is None:
                a.variables[slotName].default = op.get("value")
            else:
                for issue in a.activeIssues + a.inactiveIssues:
                    if issue.id == issueId:
                        issue.slotValues[slotName].default = op.get(
                            "value")
                        break
        else:
            print(op.type)
            raise TypeError
    return a
